@@ -1,5 +1,4 @@
-#
-#  Copyright 2009-2013 MongoDB, Inc.
+#  Copyright 2009 - present MongoDB, Inc.
 #
 #  Licensed under the Apache License, Version 2.0 (the "License");
 #  you may not use this file except in compliance with the License.
@@ -12,27 +11,27 @@
 #  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #  See the License for the specific language governing permissions and
 #  limitations under the License.
-#
 
+use strict;
+use warnings;
 package MongoDB::Cursor;
 
 
 # ABSTRACT: A lazy cursor for Mongo query results
 
 use version;
-our $VERSION = 'v1.2.2';
+our $VERSION = 'v2.0.3';
 
 use Moo;
-use MongoDB;
-use MongoDB::BSON;
 use MongoDB::Error;
 use MongoDB::QueryResult;
 use MongoDB::ReadPreference;
 use MongoDB::_Protocol;
 use MongoDB::Op::_Explain;
-use MongoDB::_Types -types, 'to_IxHash';
+use MongoDB::_Types -types, qw/to_IxHash is_OrderedDoc/;
 use Types::Standard qw(
     InstanceOf
+    is_Str
 );
 use boolean;
 use Tie::IxHash;
@@ -47,13 +46,23 @@ use namespace::clean -except => 'meta';
 #pod
 #pod =cut
 
-with 'MongoDB::Role::_Cursor';
+with $_ for qw(
+  MongoDB::Role::_CursorAPI
+  MongoDB::Role::_DeprecationWarner
+);
 
 # attributes for sending a query
-has query => (
-    is => 'ro',
-    isa => InstanceOf['MongoDB::_Query'],
+has client => (
+    is       => 'ro',
     required => 1,
+    isa      => InstanceOf ['MongoDB::MongoClient'],
+);
+
+has _query => (
+    is => 'ro',
+    isa => InstanceOf['MongoDB::Op::_Query'],
+    required => 1,
+    init_arg => 'query',
 );
 
 # lazy result attribute
@@ -68,7 +77,8 @@ has result => (
 # this does the query if it hasn't been done yet
 sub _build_result {
     my ($self) = @_;
-    $self->query->execute;
+
+    return $self->{client}->send_read_op( $self->_query );
 }
 
 #--------------------------------------------------------------------------#
@@ -105,7 +115,7 @@ sub immortal {
     MongoDB::UsageError->throw("cannot set immortal after querying")
         if $self->started_iterating;
 
-    $self->query->noCursorTimeout(!!$bool);
+    $self->_query->set_noCursorTimeout($bool);
     return $self;
 }
 
@@ -137,7 +147,7 @@ sub fields {
     MongoDB::UsageError->throw("not a hash reference")
       unless ref $f eq 'HASH' || ref $f eq 'Tie::IxHash';
 
-    $self->query->projection($f);
+    $self->_query->set_projection($f);
     return $self;
 }
 
@@ -159,7 +169,7 @@ sub sort {
     MongoDB::UsageError->throw("cannot set sort after querying")
       if $self->started_iterating;
 
-    $self->query->sort( to_IxHash($order) );
+    $self->_query->set_sort($order);
     return $self;
 }
 
@@ -178,7 +188,7 @@ sub limit {
     my ( $self, $num ) = @_;
     MongoDB::UsageError->throw("cannot set limit after querying")
       if $self->started_iterating;
-    $self->query->limit($num);
+    $self->_query->set_limit($num);
     return $self;
 }
 
@@ -204,7 +214,7 @@ sub max_await_time_ms {
     MongoDB::UsageError->throw("can not set max_await_time_ms after querying")
       if $self->started_iterating;
 
-    $self->query->maxAwaitTimeMS( $num );
+    $self->_query->set_maxAwaitTimeMS( $num );
     return $self;
 }
 
@@ -227,7 +237,7 @@ sub max_time_ms {
     MongoDB::UsageError->throw("can not set max_time_ms after querying")
       if $self->started_iterating;
 
-    $self->query->maxTimeMS( $num );
+    $self->_query->set_maxTimeMS( $num );
     return $self;
 
 }
@@ -257,7 +267,7 @@ sub tailable {
     MongoDB::UsageError->throw("cannot set tailable after querying")
         if $self->started_iterating;
 
-    $self->query->cursorType($bool ? 'tailable' : 'non_tailable');
+    $self->_query->set_cursorType($bool ? 'tailable' : 'non_tailable');
     return $self;
 }
 
@@ -281,7 +291,7 @@ sub tailable_await {
     MongoDB::UsageError->throw("cannot set tailable_await after querying")
         if $self->started_iterating;
 
-    $self->query->cursorType($bool ? 'tailable_await' : 'non_tailable');
+    $self->_query->set_cursorType($bool ? 'tailable_await' : 'non_tailable');
     return $self;
 }
 
@@ -302,45 +312,24 @@ sub skip {
     MongoDB::UsageError->throw("cannot set skip after querying")
       if $self->started_iterating;
 
-    $self->query->skip($num);
-    return $self;
-}
-
-#pod =head2 snapshot
-#pod
-#pod     $cursor->snapshot(1);
-#pod
-#pod Uses snapshot mode for the query.  Snapshot mode assures no duplicates are
-#pod returned due an intervening write relocating a document.  Note that if an
-#pod object is inserted, updated or deleted during the query, it may or may not
-#pod be returned when snapshot mode is enabled. Short query responses (less than
-#pod 1MB) are always effectively snapshotted.  Currently, snapshot mode may not
-#pod be used with sorting or explicit hints.
-#pod
-#pod Returns this cursor for chaining operations.
-#pod
-#pod =cut
-
-sub snapshot {
-    my ($self, $bool) = @_;
-
-    MongoDB::UsageError->throw("cannot set snapshot after querying")
-      if $self->started_iterating;
-
-    MongoDB::UsageError->throw("snapshot requires a defined, boolean argument")
-      unless defined $bool;
-
-    $self->query->modifiers->{'$snapshot'} = $bool;
+    $self->_query->set_skip($num);
     return $self;
 }
 
 #pod =head2 hint
 #pod
-#pod     $cursor->hint({'x' => 1});
-#pod     $cursor->hint(['x', 1]);
-#pod     $cursor->hint('x_1');
+#pod Hint the query to use a specific index by name:
 #pod
-#pod Force Mongo to use a specific index for a query.
+#pod     $cursor->hint("index_name");
+#pod
+#pod Hint the query to use index based on individual keys and direction:
+#pod
+#pod     $cursor->hint([field_1 => 1, field_2 => -1, field_3 => 1]);
+#pod
+#pod Use of a hash reference should be avoided except for single key indexes.
+#pod
+#pod The hint must be a string or L<ordered document|MongoDB::Collection/Ordered
+#pod document>.
 #pod
 #pod Returns this cursor for chaining operations.
 #pod
@@ -350,16 +339,11 @@ sub hint {
     my ( $self, $index ) = @_;
     MongoDB::UsageError->throw("cannot set hint after querying")
       if $self->started_iterating;
+    MongoDB::UsageError->throw("hint must be string or ordered document, not '$index'")
+      if ! (is_Str($index) || is_OrderedDoc($index));
 
-    # $index must either be a string or a reference to an array, hash, or IxHash
-    if ( ref $index eq 'ARRAY' ) {
-        $index = Tie::IxHash->new(@$index);
-    }
-    elsif ( ref $index && !( ref $index eq 'HASH' || ref $index eq 'Tie::IxHash' ) ) {
-        MongoDB::UsageError->throw("not a hash reference");
-    }
+    $self->_query->set_hint( $index );
 
-    $self->query->modifiers->{'$hint'} = $index;
     return $self;
 }
 
@@ -381,7 +365,7 @@ sub partial {
     MongoDB::UsageError->throw("cannot set partial after querying")
       if $self->started_iterating;
 
-    $self->query->allowPartialResults( !! $value );
+    $self->_query->set_allowPartialResults( $value );
 
     # returning self is an API change but more consistent with other cursor methods
     return $self;
@@ -411,7 +395,7 @@ sub read_preference {
 
     my $type = ref $_[0];
     if ( $type eq 'MongoDB::ReadPreference' ) {
-        $self->query->read_preference( $_[0] );
+        $self->_query->read_preference( $_[0] );
     }
     else {
         my $mode     = shift || 'primary';
@@ -420,7 +404,7 @@ sub read_preference {
             mode => $mode,
             ( $tag_sets ? ( tag_sets => $tag_sets ) : () )
         );
-        $self->query->read_preference($rp);
+        $self->_query->read_preference($rp);
     }
 
     return $self;
@@ -448,15 +432,18 @@ sub explain {
     my ($self) = @_;
 
     my $explain_op = MongoDB::Op::_Explain->_new(
-        db_name         => $self->query->db_name,
-        coll_name       => $self->query->coll_name,
-        bson_codec      => $self->query->bson_codec,
-        query           => $self->query->clone,
-        read_preference => $self->query->read_preference,
-        read_concern    => $self->query->read_concern,
+        db_name             => $self->_query->db_name,
+        coll_name           => $self->_query->coll_name,
+        full_name           => $self->_query->full_name,
+        bson_codec          => $self->_query->bson_codec,
+        query               => $self->_query,
+        read_preference     => $self->_query->read_preference,
+        read_concern        => $self->_query->read_concern,
+        session             => $self->_query->session,
+        monitoring_callback => $self->client->monitoring_callback,
     );
 
-    return $self->query->client->send_read_op($explain_op);
+    return $self->_query->client->send_read_op($explain_op);
 }
 
 #pod =head1 QUERY ITERATION
@@ -591,59 +578,22 @@ sub info {
 # Deprecated methods
 #--------------------------------------------------------------------------#
 
-sub count {
-    my ($self, $limit_skip) = @_;
+sub snapshot {
+    my ($self, $bool) = @_;
 
-    my $cmd = new Tie::IxHash(count => $self->query->coll_name);
+    $self->_warn_deprecated_method(
+        'snapshot' => "Snapshot is deprecated as of MongoDB 3.6" );
 
-    $cmd->Push(query => $self->query->filter);
-
-    if ($limit_skip) {
-        $cmd->Push(limit => $self->query->limit) if $self->query->limit;
-        $cmd->Push(skip => $self->query->skip) if $self->query->skip;
-    }
-
-    if (my $hint = $self->query->modifiers->{'$hint'}) {
-        $cmd->Push(hint => $hint);
-    }
-
-    my $result = try {
-        my $db = $self->query->client->get_database( $self->query->db_name );
-        $db->run_command( $cmd, $self->query->read_preference );
-    }
-    catch {
-        # if there was an error, check if it was the "ns missing" one that means the
-        # collection hasn't been created or a real error.
-        die $_ unless /^ns missing/;
-    };
-
-    return $result ? $result->{n} : 0;
-}
-
-my $PRIMARY = MongoDB::ReadPreference->new;
-my $SEC_PREFERRED = MongoDB::ReadPreference->new( mode => 'secondaryPreferred' );
-
-sub slave_okay {
-    my ($self, $value) = @_;
-    MongoDB::UsageError->throw("cannot set slave_ok after querying")
+    MongoDB::UsageError->throw("cannot set snapshot after querying")
       if $self->started_iterating;
 
-    if ($value) {
-        # if not 'primary', then slave_ok is already true, so leave alone
-        if ( $self->query->read_preference->mode eq 'primary' ) {
-            # secondaryPreferred is how mongos interpretes slave_ok
-            $self->query->read_preference( $SEC_PREFERRED );
-        }
-    }
-    else {
-        $self->query->read_preference( $PRIMARY );
-    }
+    MongoDB::UsageError->throw("snapshot requires a defined, boolean argument")
+      unless defined $bool;
 
-    # returning self is an API change but more consistent with other cursor methods
+    $self->_query->set_snapshot($bool);
     return $self;
 }
 
-
 1;
 
 
@@ -662,7 +612,7 @@ MongoDB::Cursor - A lazy cursor for Mongo query results
 
 =head1 VERSION
 
-version v1.2.2
+version v2.0.3
 
 =head1 SYNOPSIS
 
@@ -808,26 +758,20 @@ Skips the first N results.
 
 Returns this cursor for chaining operations.
 
-=head2 snapshot
-
-    $cursor->snapshot(1);
-
-Uses snapshot mode for the query.  Snapshot mode assures no duplicates are
-returned due an intervening write relocating a document.  Note that if an
-object is inserted, updated or deleted during the query, it may or may not
-be returned when snapshot mode is enabled. Short query responses (less than
-1MB) are always effectively snapshotted.  Currently, snapshot mode may not
-be used with sorting or explicit hints.
-
-Returns this cursor for chaining operations.
-
 =head2 hint
 
-    $cursor->hint({'x' => 1});
-    $cursor->hint(['x', 1]);
-    $cursor->hint('x_1');
+Hint the query to use a specific index by name:
 
-Force Mongo to use a specific index for a query.
+    $cursor->hint("index_name");
+
+Hint the query to use index based on individual keys and direction:
+
+    $cursor->hint([field_1 => 1, field_2 => -1, field_3 => 1]);
+
+Use of a hash reference should be avoided except for single key indexes.
+
+The hint must be a string or L<ordered document|MongoDB::Collection/Ordered
+document>.
 
 Returns this cursor for chaining operations.
 
@@ -998,6 +942,10 @@ David Golden <david@mongodb.com>
 
 =item *
 
+Rassi <rassi@mongodb.com>
+
+=item *
+
 Mike Friedman <friedo@friedo.com>
 
 =item *
@@ -1012,7 +960,7 @@ Florian Ragwitz <rafl@debian.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is Copyright (c) 2016 by MongoDB, Inc..
+This software is Copyright (c) 2019 by MongoDB, Inc.
 
 This is free software, licensed under:
 
