@@ -1,5 +1,4 @@
-#
-#  Copyright 2014 MongoDB, Inc.
+#  Copyright 2014 - present MongoDB, Inc.
 #
 #  Licensed under the Apache License, Version 2.0 (the "License");
 #  you may not use this file except in compliance with the License.
@@ -12,21 +11,24 @@
 #  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #  See the License for the specific language governing permissions and
 #  limitations under the License.
-#
 
+use strict;
+use warnings;
 package MongoDB::Role::_WriteResult;
 
 # MongoDB interface for common write result attributes and methods
 
 use version;
-our $VERSION = 'v1.4.5';
+our $VERSION = 'v2.0.3';
+
+use Moo::Role;
 
 use MongoDB::Error;
 use MongoDB::_Constants;
 use MongoDB::_Types qw(
     ArrayOfHashRef
 );
-use Moo::Role;
+
 use namespace::clean;
 
 has [qw/write_errors write_concern_errors/] => (
@@ -35,7 +37,7 @@ has [qw/write_errors write_concern_errors/] => (
     isa      => ArrayOfHashRef,
 );
 
-with 'MongoDB::Role::_LastError';
+with 'MongoDB::Role::_DatabaseErrorThrower';
 
 sub acknowledged() { 1 }; # override to 0 for MongoDB::UnacknowledgedResult
 
@@ -120,4 +122,15 @@ sub last_wtimeout {
     return !!( $self->count_write_concern_errors && !$self->count_write_errors );
 }
 
+sub last_error_labels {
+    my ( $self ) = @_;
+    if ( $self->count_write_errors ) {
+        return $self->write_errors->[-1]{errorLabels} || [];
+    }
+    elsif ( $self->count_write_concern_errors ) {
+        return $self->write_errors->[-1]{errorLabels} || [];
+    }
+    return [];
+}
+
 1;
