@@ -1,5 +1,4 @@
-#
-#  Copyright 2014 MongoDB, Inc.
+#  Copyright 2014 - present MongoDB, Inc.
 #
 #  Licensed under the Apache License, Version 2.0 (the "License");
 #  you may not use this file except in compliance with the License.
@@ -12,36 +11,45 @@
 #  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #  See the License for the specific language governing permissions and
 #  limitations under the License.
-#
 
+use strict;
+use warnings;
 package MongoDB::_Topology;
 
 use version;
-our $VERSION = 'v1.2.2';
+our $VERSION = 'v2.0.3';
 
 use Moo;
-use MongoDB::BSON;
+use BSON;
 use MongoDB::Error;
 use MongoDB::Op::_Command;
+use MongoDB::_Platform;
 use MongoDB::ReadPreference;
 use MongoDB::_Constants;
 use MongoDB::_Link;
 use MongoDB::_Types qw(
+    Boolish
     BSONCodec
+    CompressionType
+    Document
     NonNegNum
     TopologyType
+    ZlibCompressionLevel
+    to_IxHash
 );
 use Types::Standard qw(
-    Bool
+    CodeRef
     HashRef
+    ArrayRef
     InstanceOf
     Num
     Str
     Maybe
 );
 use MongoDB::_Server;
+use MongoDB::_Protocol;
 use Config;
-use List::Util qw/first/;
+use List::Util qw/first max min/;
 use Safe::Isa;
 use Time::HiRes qw/time usleep/;
 use Try::Tiny;
@@ -58,6 +66,12 @@ has uri => (
     isa => InstanceOf['MongoDB::_URI'],
 );
 
+has min_server_version => (
+    is       => 'ro',
+    required => 1,
+    isa => Str,
+);
+
 has max_wire_version => (
     is       => 'ro',
     required => 1,
@@ -76,6 +90,26 @@ has credential => (
     isa => InstanceOf['MongoDB::_Credential'],
 );
 
+# Required so it's passed explicitly, even if undef, to ensure it's wired
+# up correctly.
+has monitoring_callback => (
+    is => 'ro',
+    required => 1,
+    isa => Maybe[CodeRef],
+);
+
+has compressors => (
+    is => 'ro',
+    isa => ArrayRef[CompressionType],
+    default => sub { [] },
+);
+
+has zlib_compression_level => (
+    is => 'ro',
+    isa => ZlibCompressionLevel,
+    default => sub { -1 },
+);
+
 has type => (
     is      => 'ro',
     writer  => '_set_type',
@@ -83,6 +117,12 @@ has type => (
     isa => TopologyType,
 );
 
+has app_name => (
+    is      => 'ro',
+    default => '',
+    isa     => Str,
+);
+
 has replica_set_name => (
     is      => 'ro',
     default => '',
@@ -109,6 +149,20 @@ has local_threshold_sec => (
     isa => Num,
 );
 
+has logical_session_timeout_minutes => (
+    is => 'ro',
+    default => undef,
+    writer => '_set_logical_session_timeout_minutes',
+    isa => Maybe [NonNegNum],
+);
+
+has next_scan_time => (
+    is      => 'ro',
+    default => sub { time() },
+    writer  => '_set_next_scan_time',
+    isa => Num,
+);
+
 has socket_check_interval_sec => (
     is      => 'ro',
     default => 5,
@@ -124,7 +178,7 @@ has server_selection_timeout_sec => (
 has server_selection_try_once => (
     is      => 'ro',
     default => 1,
-    isa => Bool,
+    isa => Boolish,
 );
 
 has ewma_alpha => (
@@ -141,8 +195,9 @@ has link_options => (
 
 has bson_codec => (
     is       => 'ro',
-    default  => sub { MongoDB::BSON->new },
+    default  => sub { BSON->new },
     isa => BSONCodec,
+    init_arg => undef,
 );
 
 has number_of_seeds => (
@@ -153,7 +208,6 @@ has number_of_seeds => (
 
 has max_election_id => (
     is      => 'rw',
-    isa     => Maybe[ InstanceOf['MongoDB::OID'] ],
     writer  => '_set_max_election_id',
 );
 
@@ -163,11 +217,30 @@ has max_set_version => (
     writer => '_set_max_set_version',
 );
 
+# generated only once per _Topology, for performance
+has handshake_document => (
+    is      => 'lazy',
+    isa     => Document,
+    builder => '_build_handshake_document',
+);
+
 # compatible wire protocol
 has is_compatible => (
     is => 'ro',
     writer => '_set_is_compatible',
-    isa => Bool,
+    isa => Boolish,
+);
+
+has wire_version_floor => (
+    is => 'ro',
+    writer => '_set_wire_version_floor',
+    default => 0,
+);
+
+has wire_version_ceil => (
+    is => 'ro',
+    writer => '_set_wire_version_ceil',
+    default => 0,
 );
 
 has current_primary => (
@@ -190,6 +263,12 @@ has servers => (
     isa => HashRef[InstanceOf['MongoDB::_Server']],
 );
 
+has _incompatible_servers => (
+    is => 'rw',
+    default => sub { [] },
+    isa => ArrayRef[InstanceOf['MongoDB::_Server']],
+);
+
 has links => (
     is      => 'ro',
     default => sub { {} },
@@ -202,19 +281,69 @@ has rtt_ewma_sec => (
     isa => HashRef[Num],
 );
 
+has cluster_time => (
+    is => 'rwp',
+    isa => Maybe[Document],
+    init_arg => undef,
+    default => undef,
+);
+
+sub update_cluster_time {
+    my ( $self, $cluster_time ) = @_;
+
+    return unless $cluster_time && exists $cluster_time->{clusterTime}
+        && ref($cluster_time->{clusterTime}) eq 'BSON::Timestamp';
+
+    # Only update the cluster time if it is more recent than the current entry
+    if ( !defined $self->cluster_time ) {
+        $self->_set_cluster_time($cluster_time);
+    }
+    elsif ( $cluster_time->{'clusterTime'} > $self->cluster_time->{'clusterTime'} ) {
+        $self->_set_cluster_time($cluster_time);
+    }
+    return;
+}
+
 #--------------------------------------------------------------------------#
 # builders
 #--------------------------------------------------------------------------#
 
 sub _build_number_of_seeds {
     my ($self) = @_;
-    return scalar @{ $self->uri->hostpairs };
+    return scalar @{ $self->uri->hostids };
+}
+
+sub _truncate_for_handshake {
+    my $str = shift;
+    return substr( $str, 0, 64 );
+}
+
+sub _build_handshake_document {
+    my ($self) = @_;
+    my $driver_version_without_leading_v = substr( $VERSION, 1 );
+
+    return to_IxHash(
+        [
+            ( length( $self->app_name ) ? ( application => { name => $self->app_name } ) : () ),
+            driver => to_IxHash(
+                [
+                    name    => "MongoDB Perl Driver",
+                    version => $driver_version_without_leading_v,
+                ]
+            ),
+            os => { type => _truncate_for_handshake(MongoDB::_Platform::os_type) },
+            platform => _truncate_for_handshake(MongoDB::_Platform::platform_details)
+        ]
+    );
 }
 
 sub BUILD {
     my ($self) = @_;
     my $type = $self->type;
-    my @addresses = @{ $self->uri->hostpairs };
+    my @addresses = @{ $self->uri->hostids };
+
+    # clone bson codec to disable dt_type
+    $self->{bson_codec} = $self->bson_codec->clone( dt_type => undef );
 
     if ( my $set_name = $self->replica_set_name ) {
         if ( $type eq 'Single' || $type eq 'ReplicaSetNoPrimary' ) {
@@ -245,12 +374,14 @@ sub BUILD {
 
 sub all_servers { return values %{ $_[0]->servers } }
 
+sub all_data_bearing_servers { return grep { $_->is_data_bearing } $_[0]->all_servers }
+
 sub check_address {
     my ( $self, $address ) = @_;
 
     my $link = $self->links->{$address};
     if ( $link && $link->is_connected ) {
-        $self->_update_topology_from_link($link);
+        $self->_update_topology_from_link( $link, with_handshake => 0 );
     }
     else {
         # initialize_link will call update_topology_from_link
@@ -271,11 +402,12 @@ sub get_readable_link {
 
     my $mode = $read_pref ? lc $read_pref->mode : 'primary';
     my $method =
-      ( $self->type eq "Single" || $self->type eq "Sharded" )
-      ? '_find_any_server'
-      : "_find_${mode}_server";
+        $self->type eq "Single"  ? '_find_available_server'
+      : $self->type eq "Sharded" ? '_find_readable_mongos_server'
+      :                            "_find_${mode}_server";
 
-    if ($mode eq 'primary' && $self->current_primary) {
+    if ( $mode eq 'primary' && $self->current_primary && $self->next_scan_time > time() )
+    {
         my $link = $self->_get_server_link( $self->current_primary, $method );
         return $link if $link;
     }
@@ -314,11 +446,11 @@ sub get_writable_link {
 
     my $method =
       ( $self->type eq "Single" || $self->type eq "Sharded" )
-      ? '_find_any_server'
+      ? '_find_available_server'
       : "_find_primary_server";
 
 
-    if ($self->current_primary) {
+    if ( $self->current_primary && $self->next_scan_time > time() ) {
         my $link = $self->_get_server_link( $self->current_primary, $method );
         return $link if $link;
     }
@@ -350,11 +482,11 @@ sub mark_stale {
 }
 
 sub scan_all_servers {
-    my ($self) = @_;
+    my ($self, $force) = @_;
 
     my ( $next, @ordinary, @to_check );
     my $start_time = time;
-    my $cooldown_time = $start_time - COOLDOWN_SECS;
+    my $cooldown_time = $force ? $start_time : $start_time - COOLDOWN_SECS;
 
     # anything not updated since scan start is eligible for a check; when all servers
     # are updated, the loop terminates; Unknown servers aren't checked if
@@ -384,7 +516,9 @@ sub scan_all_servers {
         }
     }
 
-    $self->_set_last_scan_time( time );
+    my $now = time();
+    $self->_set_last_scan_time( $now );
+    $self->_set_next_scan_time( $now + $self->heartbeat_frequency_sec );
     $self->_set_stale( 0 );
     $self->_check_wire_versions;
     return;
@@ -451,21 +585,115 @@ sub _check_oldest_server {
     return;
 }
 
+my $max_int32 = 2147483647;
+
 sub _check_wire_versions {
     my ($self) = @_;
 
     my $compat = 1;
+    my $min_seen = $max_int32;
+    my $max_seen = 0;
     for my $server ( grep { $_->is_available } $self->all_servers ) {
         my ( $server_min_wire_version, $server_max_wire_version ) =
           @{ $server->is_master }{qw/minWireVersion maxWireVersion/};
 
-        if (   ( $server_min_wire_version || 0 ) > $self->max_wire_version
-            || ( $server_max_wire_version || 0 ) < $self->min_wire_version )
-        {
+        # set to 0 as could be undefined. 0 is the equivalent to missing, and
+        # also kept as 0 for legacy compatibility.
+        $server_max_wire_version = 0 unless defined $server_max_wire_version;
+        $server_min_wire_version = 0 unless defined $server_min_wire_version;
+
+        if ( $server_min_wire_version > $self->max_wire_version
+          || $server_max_wire_version < $self->min_wire_version ) {
             $compat = 0;
+            push @{ $self->_incompatible_servers }, $server;
         }
+
+        $min_seen = $server_max_wire_version if $server_max_wire_version < $min_seen;
+        $max_seen = $server_max_wire_version if $server_max_wire_version > $max_seen;
     }
     $self->_set_is_compatible($compat);
+    $self->_set_wire_version_floor($min_seen);
+    $self->_set_wire_version_ceil($max_seen);
+
+    return;
+}
+
+sub _update_ls_timeout_minutes {
+    my ( $self, $new_server ) = @_;
+
+    my @data_bearing_servers = grep { $_->is_data_bearing } $self->all_servers;
+    my $timeout = min map {
+        # use -1 as a flag to prevent undefined warnings
+        defined $_->logical_session_timeout_minutes
+          ? $_->logical_session_timeout_minutes
+          : -1
+    } @data_bearing_servers;
+    # min will return undef if the array is empty
+    if ( defined $timeout && $timeout < 0 ) {
+        $timeout = undef;
+    }
+    $self->_set_logical_session_timeout_minutes( $timeout );
+    return;
+}
+
+sub _supports_sessions {
+    my ( $self ) = @_;
+
+    $self->scan_all_servers if $self->stale;
+
+    my @servers = $self->all_servers;
+    return if @servers == 1 && $servers[0]->type eq 'Standalone';
+
+    return defined $self->logical_session_timeout_minutes;
+}
+
+# Used for bulkWrite for shortcutting to original execute command
+sub _supports_retry_writes {
+    my ( $self ) = @_;
+
+    # retryWrites arent supported in standalone servers
+    return if $self->type eq 'Single';
+
+    # retryWrites require a wire version of at least six
+    return if $self->wire_version_ceil < 6;
+
+    # must have lstm present
+    return 1 if defined $self->logical_session_timeout_minutes;
+    return;
+}
+
+sub _supports_transactions {
+    my ( $self ) = @_;
+
+    return unless $self->_supports_sessions;
+    return if $self->wire_version_ceil < 7;
+    return if $self->type eq 'Sharded';
+    return 1;
+}
+
+sub _check_staleness_compatibility {
+    my ($self, $read_pref) = @_;
+    my $max_staleness_sec = $read_pref ? $read_pref->max_staleness_seconds : -1;
+
+    if ( $max_staleness_sec > 0 ) {
+        if ( $self->wire_version_floor < 5 ) {
+            MongoDB::ProtocolError->throw(
+                "Incompatible wire protocol version. You tried to use max_staleness_seconds with one or more servers that don't support it."
+            );
+        }
+
+        if (
+            ( $self->type eq "ReplicaSetWithPrimary" || $self->type eq "ReplicaSetNoPrimary" )
+            && $max_staleness_sec < max( SMALLEST_MAX_STALENESS_SEC,
+                $self->heartbeat_frequency_sec + IDLE_WRITE_PERIOD_SEC
+            )
+          )
+        {
+            MongoDB::UsageError->throw(
+                "max_staleness_seconds must be at least 90 seconds and at least heartbeat_frequency (in secs) + 10 secs"
+            );
+        }
+    }
 
     return;
 }
@@ -478,29 +706,89 @@ sub _dump {
 sub _eligible {
     my ( $self, $read_pref, @candidates ) = @_;
 
-    return @candidates
-      if $read_pref->has_empty_tag_sets;
+    # must filter on max staleness first, so that the remaining servers
+    # are checked against the list of tag_sets
+    if ( $read_pref->max_staleness_seconds > 0 ) {
+        @candidates = $self->_filter_fresh_servers($read_pref, @candidates );
+        return unless @candidates;
+    };
 
     # given a tag set list, if a tag set matches at least one
     # candidate, then all candidates matching that tag set are eligible
-    for my $ts ( @{ $read_pref->tag_sets } ) {
-        my @eligible = grep { $_->matches_tag_set($ts) } @candidates;
-        return @eligible if @eligible;
+    if ( ! $read_pref->has_empty_tag_sets ) {
+        for my $ts ( @{ $read_pref->tag_sets } ) {
+            if ( my @ts_candidates = grep { $_->matches_tag_set($ts) } @candidates ) {
+                return @ts_candidates;
+            }
+        }
+        return;
     }
 
-    return;
+    return @candidates;
 }
 
-sub _find_any_server {
-    my ( $self, undef, @candidates ) = @_;
+sub _filter_fresh_servers {
+    my ($self, $read_pref, @candidates) = @_;
+
+    # all values should be floating point seconds
+    my $max_staleness_sec = $read_pref->max_staleness_seconds;
+    my $heartbeat_frequency_sec = $self->heartbeat_frequency_sec;
+
+    if ( $self->type eq 'ReplicaSetWithPrimary' ) {
+        my ($primary) = $self->_primaries;
+
+        # all values should be floating point seconds
+        my $p_last_write_date = $primary->last_write_date;
+        my $p_last_update_time = $primary->last_update_time;
+
+        return map { $_->[0] }
+          grep { $_->[1] <= $max_staleness_sec }
+          map {
+            [
+                $_,
+                $p_last_write_date
+                  + ( $_->last_update_time - $p_last_update_time )
+                  - $_->last_write_date
+                  + $heartbeat_frequency_sec
+            ]
+          } @candidates;
+    }
+    else {
+        my ($smax) = map { $_->[0] }
+          sort { $b->[1] <=> $a->[1] }
+          map { [ $_, $_->last_write_date ] } $self->_secondaries;
+        my $smax_last_write_date = $smax->last_write_date;
+
+        return map { $_->[0] }
+          grep     { $_->[1] <= $max_staleness_sec }
+          map {
+            [ $_, $smax_last_write_date - $_->last_write_date + $heartbeat_frequency_sec ]
+          } @candidates;
+    }
+}
+
+# This works for reads and writes; for writes, $read_pref will be undef
+sub _find_available_server {
+    my ( $self, $read_pref, @candidates ) = @_;
+    $self->_check_staleness_compatibility($read_pref) if $read_pref;
     push @candidates, $self->all_servers unless @candidates;
     return $self->_get_server_in_latency_window(
         [ grep { $_->is_available } @candidates ] );
 }
 
+# This uses read preference to check for max staleness compatibility in
+# mongos, but otherwise read preference is ignored (mongos will pass it on)
+sub _find_readable_mongos_server {
+    my ( $self, $read_pref, @candidates ) = @_;
+    $self->_check_staleness_compatibility($read_pref);
+    push @candidates, $self->all_servers unless @candidates;
+    return $self->_get_server_in_latency_window(
+        [ grep { $_->is_available } @candidates ] );
+}
 
 sub _find_nearest_server {
     my ( $self, $read_pref, @candidates ) = @_;
+    $self->_check_staleness_compatibility($read_pref);
     push @candidates, ( $self->_primaries, $self->_secondaries ) unless @candidates;
     my @suitable = $self->_eligible( $read_pref, @candidates );
     return $self->_get_server_in_latency_window( \@suitable );
@@ -516,12 +804,14 @@ sub _find_primary_server {
 
 sub _find_primarypreferred_server {
     my ( $self, $read_pref, @candidates ) = @_;
+    $self->_check_staleness_compatibility($read_pref);
     return $self->_find_primary_server(@candidates)
       || $self->_find_secondary_server( $read_pref, @candidates );
 }
 
 sub _find_secondary_server {
     my ( $self, $read_pref, @candidates ) = @_;
+    $self->_check_staleness_compatibility($read_pref);
     push @candidates, $self->_secondaries unless @candidates;
     my @suitable = $self->_eligible( $read_pref, @candidates );
     return $self->_get_server_in_latency_window( \@suitable );
@@ -529,6 +819,7 @@ sub _find_secondary_server {
 
 sub _find_secondarypreferred_server {
     my ( $self, $read_pref, @candidates ) = @_;
+    $self->_check_staleness_compatibility($read_pref);
     return $self->_find_secondary_server( $read_pref, @candidates )
       || $self->_find_primary_server(@candidates);
 }
@@ -563,7 +854,7 @@ sub _get_server_link {
     return unless $link;
 
     # for idle links, refresh the server and verify validity
-    if ( $link->idle_time_sec > $self->socket_check_interval_sec ) {
+    if ( time - $link->last_used > $self->socket_check_interval_sec ) {
         $self->check_address($address);
 
         # topology might have dropped the server
@@ -596,7 +887,7 @@ sub _initialize_link {
 
     # connection succeeded, so register link and get a server description
     $self->links->{$address} = $link;
-    $self->_update_topology_from_link($link);
+    $self->_update_topology_from_link( $link, with_handshake => 1 );
 
     # after update, server might or might not exist in the topology;
     # if not, return nothing
@@ -605,14 +896,15 @@ sub _initialize_link {
     # we have a link and the server is a valid member, so
     # try to authenticate; if authentication fails, all
     # servers are considered invalid and we throw an error
-    if ( first { $_ eq $server->type } qw/Standalone Mongos RSPrimary RSSecondary/ ) {
+    if ( $self->type eq 'Single' || first { $_ eq $server->type } qw/Standalone Mongos RSPrimary RSSecondary/ ) {
         try {
-            $self->credential->authenticate($link, $self->bson_codec);
+            $self->credential->authenticate($server, $link, $self->bson_codec);
         }
         catch {
             my $err = $_;
+            my $msg = $err->$_isa("MongoDB::Error") ? $err->message : "$err";
             $self->_reset_address_to_unknown( $_->address, $err ) for $self->all_servers;
-            MongoDB::AuthError->throw("Authentication to $address failed: $err");
+            MongoDB::AuthError->throw("Authentication to $address failed: $msg");
         };
     }
 
@@ -676,7 +968,7 @@ sub _selection_timeout {
     my $start_time = my $loop_end_time = time();
     my $max_time = $start_time + $self->server_selection_timeout_sec;
 
-    if ( $self->last_scan_time + $self->heartbeat_frequency_sec < $start_time ) {
+    if ( $self->next_scan_time < $start_time ) {
         $self->_set_stale(1);
     }
 
@@ -697,9 +989,31 @@ sub _selection_timeout {
 
         unless ( $self->is_compatible ) {
             $self->_set_stale(1);
-            MongoDB::ProtocolError->throw(
-                "Incompatible wire protocol version. This version of the MongoDB driver is not compatible with the server. You probably need to upgrade this library."
-            );
+            my $error_string = '';
+            for my $server ( @{ $self->_incompatible_servers } ) {
+                my $min_wire_ver = $server->is_master->{minWireVersion};
+                my $max_wire_ver = $server->is_master->{maxWireVersion};
+                my $host         = $server->address;
+                if ( $min_wire_ver > $self->max_wire_version ) {
+                    $error_string .= sprintf(
+                        "Server at %s requires wire version %d, but this version of %s only supports up to %d.\n",
+                        $host,
+                        $min_wire_ver,
+                        'Perl MongoDB Driver',
+                        $self->max_wire_version
+                    );
+                } else {
+                    $error_string .= sprintf(
+                        "Server at %s reports wire version %d, but this version of %s requires at least %d (MongoDB %s).\n",
+                        $host,
+                        $max_wire_ver,
+                        'Perl MongoDB Driver',
+                        $self->min_wire_version,
+                        $self->min_server_version,
+                    );
+                }
+            }
+            MongoDB::ProtocolError->throw( $error_string );
         }
 
         my $server = $self->$method($read_pref);
@@ -722,17 +1036,38 @@ sub _selection_timeout {
 
 my $PRIMARY = MongoDB::ReadPreference->new;
 
+sub _generate_ismaster_request {
+    my ( $self, $link, $should_perform_handshake ) = @_;
+    my @opts;
+    if ($should_perform_handshake) {
+        push @opts, client => $self->handshake_document;
+        if ( $self->credential->mechanism eq 'DEFAULT' ) {
+            my $db_user = join( ".", map { $self->credential->$_ } qw/source username/ );
+            push @opts, saslSupportedMechs => $db_user;
+        }
+        if (@{ $self->compressors }) {
+            push @opts, compression => $self->compressors;
+        }
+    }
+    if ( $link->supports_clusterTime && defined $self->cluster_time ) {
+        push @opts, '$clusterTime' => $self->cluster_time;
+    }
+
+    return [ ismaster => 1, @opts ];
+}
+
 sub _update_topology_from_link {
-    my ( $self, $link ) = @_;
+    my ( $self, $link, %opts ) = @_;
 
     my $start_time = time;
     my $is_master = eval {
         my $op = MongoDB::Op::_Command->_new(
-            db_name         => 'admin',
-            query           => [ ismaster => 1 ],
-            query_flags     => {},
-            bson_codec      => $self->bson_codec,
-            read_preference => $PRIMARY,
+            db_name             => 'admin',
+            query               => $self->_generate_ismaster_request( $link, $opts{with_handshake} ),
+            query_flags         => {},
+            bson_codec          => $self->bson_codec,
+            read_preference     => $PRIMARY,
+            monitoring_callback => $self->monitoring_callback,
         );
         # just for this command, use connect timeout as socket timeout;
         # this violates encapsulation, but requires less API modification
@@ -740,13 +1075,15 @@ sub _update_topology_from_link {
         local $link->{socket_timeout} = $link->{connect_timeout};
         $op->execute( $link )->output;
     };
-    if ( $@ ) {
-        local $_ = $@;
-        warn "During MongoDB topology update for @{[$link->address]}: $_"
+    if ( my $e = $@ ) {
+        if ($e->$_isa("MongoDB::DatabaseError") && $e->code == USER_NOT_FOUND ) {
+            MongoDB::AuthError->throw("mechanism negotiation error: $e");
+        }
+        warn "During MongoDB topology update for @{[$link->address]}: $e"
             if WITH_ASSERTS;
-        $self->_reset_address_to_unknown( $link->address, $_ );
+        $self->_reset_address_to_unknown( $link->address, $e );
         # retry a network error if server was previously known to us
-        if (    $_->$_isa("MongoDB::NetworkError")
+        if (    $e->$_isa("MongoDB::NetworkError")
             and $link->server
             and $link->server->type ne 'Unknown'
             and $link->server->type ne 'PossiblePrimary' )
@@ -760,6 +1097,10 @@ sub _update_topology_from_link {
 
     return unless $is_master;
 
+    if ( my $cluster_time = $is_master->{'$clusterTime'} ) {
+        $self->update_cluster_time($cluster_time);
+    }
+
     my $end_time = time;
     my $rtt_sec = $end_time - $start_time;
 
@@ -768,6 +1109,7 @@ sub _update_topology_from_link {
         last_update_time => $end_time,
         rtt_sec           => $rtt_sec,
         is_master        => $is_master,
+        compressor       => $self->_construct_compressor($is_master),
     );
 
     $self->_update_topology_from_server_desc( $link->address, $new_server );
@@ -775,6 +1117,28 @@ sub _update_topology_from_link {
     return;
 }
 
+# find suitable compressor
+#
+# implemented here because the result is based on the specified
+# order of compressors combined with the list of server supported
+# compressors
+sub _construct_compressor {
+    my ($self, $is_master) = @_;
+
+    my @supported = @{ ($is_master || {})->{compression} || [] }
+        or return undef; ## no critic
+
+    for my $name (@{ $self->compressors }) {
+        if (grep { $name eq $_ } @supported) {
+            return MongoDB::_Protocol::get_compressor($name, {
+                zlib_compression_level => $self->zlib_compression_level,
+            });
+        }
+    }
+
+    return undef; ## no critic
+}
+
 sub _update_topology_from_server_desc {
     my ( $self, $address, $new_server ) = @_;
 
@@ -795,6 +1159,8 @@ sub _update_topology_from_server_desc {
     # if link is still around, tag it with server specifics
     $self->_update_link_metadata( $address, $new_server );
 
+    $self->_update_ls_timeout_minutes( $new_server );
+
     return $new_server;
 }
 
@@ -836,8 +1202,9 @@ sub _update_rs_with_primary_from_member {
         $self->_remove_server($new_server);
     }
 
-    # require 'me' that matches expected address
-    if ( $new_server->me && $new_server->me ne $new_server->address ) {
+    # require 'me' that matches expected address.
+    # check is case insensitive
+    if ( $new_server->me && lc $new_server->me ne $new_server->address ) {
         $self->_remove_server($new_server);
         $self->_check_for_primary;
         return;
@@ -883,7 +1250,7 @@ sub _update_rs_with_primary_from_primary {
             && (
                 $max_set_version > $set_version
                 || (   $max_set_version == $set_version
-                    && $max_election_id->value gt $election_id->value )
+                    && "$max_election_id" gt "$election_id" )
             )
           )
         {
@@ -1039,26 +1406,46 @@ sub _update_Single {
     return;
 }
 
+# Direct mode is like Unknown, except that it switches only between Sharded
+# or Single based on the response.
+sub _update_Direct {
+    my ( $self, $address, $new_server ) = @_;
+
+    my $server_type = $new_server->type;
+
+    if ( $server_type eq 'Mongos' ) {
+        $self->_set_type('Sharded');
+        return;
+    }
+
+    $self->_set_type('Single');
+    return;
+}
+
 sub _update_Unknown {
     my ( $self, $address, $new_server ) = @_;
 
     my $server_type = $new_server->type;
 
+    # Starting from topology type 'unknown', a standalone server when we
+    # were given multiple seeds must be a replica set member in maintenance
+    # mode so we drop it and will rediscover it later.
     if ( $server_type eq 'Standalone' ) {
-        if ( $self->number_of_seeds == 1 ) {
-            $self->_set_type('Single');
-        }
-        else {
-            # a standalone server with multiple seeds is a replica set member
-            # in maintenance mode; we drop it and may pick it up later if it
-            # rejoins the replica set.
+        if ( $self->number_of_seeds > 1 ) {
             $self->_remove_address($address);
         }
+        else {
+            $self->_set_type('Single');
+        }
+        return;
     }
-    elsif ( $server_type eq 'Mongos' ) {
+
+    if ( $server_type eq 'Mongos' ) {
         $self->_set_type('Sharded');
+        return;
     }
-    elsif ( $server_type eq 'RSPrimary' ) {
+
+    if ( $server_type eq 'RSPrimary' ) {
         $self->_set_type('ReplicaSetWithPrimary');
         $self->_update_rs_with_primary_from_primary($new_server);
         # topology changes might have removed all primaries
